package speiger.src.coreengine.rendering.gui.font;

import speiger.src.coreengine.assets.base.IAssetProvider;
import speiger.src.coreengine.rendering.gui.font.glyth.Glyth;
import speiger.src.coreengine.rendering.gui.font.glyth.IGlythSheetInfo;
import speiger.src.coreengine.rendering.textures.base.BaseTexture;
import speiger.src.coreengine.rendering.textures.base.TextureMetadata;
import speiger.src.coreengine.rendering.utils.GLFunctions;
import speiger.src.coreengine.rendering.utils.values.GLDataType;
import speiger.src.coreengine.rendering.utils.values.IGLValue.ITextureComponentFormat;
import speiger.src.coreengine.rendering.utils.values.textures.GLTextureFormat;
import speiger.src.coreengine.rendering.utils.values.textures.GLTextureParameter;
import speiger.src.coreengine.rendering.utils.values.textures.GLTextureValue;

public class FontTexture extends BaseTexture {
	private static final TextureMetadata COLOR = TextureMetadata.builder()
			.externalFormat(GLTextureFormat.RGBA).internalFormat(GLTextureFormat.RGBA).dataFormat(GLDataType.UNSIGNED_BYTE).mipmapping(false)
			.arguement(GLTextureParameter.MIN_FILTER, GLTextureValue.NEAREST).arguement(GLTextureParameter.MAG_FILTER, GLTextureValue.NEAREST)
			.arguement(GLTextureParameter.WRAP_S, GLTextureValue.CLAMP_TO_EDGE).arguement(GLTextureParameter.WRAP_T, GLTextureValue.CLAMP_TO_EDGE).build();
	private static final TextureMetadata NO_COLOR = COLOR.copy().internalFormat(GLTextureFormat.LUMINANCE).externalFormat(GLTextureFormat.LUMINANCE).arguement(GLTextureParameter.SWIZZLE_RGBA, GLTextureValue.RED).build();
	
	boolean color;
	int bounds;
	Slot slot;
	
	public FontTexture(int bounds, boolean color) {
		this.bounds = bounds;
		this.color = color;
		slot = new Slot(0, 0, bounds, bounds);
		track();
	}
	
	public static ITextureComponentFormat formatByColor(boolean colored) {
		return colored ? GLTextureFormat.RGBA : GLTextureFormat.LUMINANCE;
	}
	
	public static TextureMetadata metadataByColor(boolean colored) {
		return colored ? COLOR : NO_COLOR;
	}
	
	@Override
	public void load(IAssetProvider provider) {
		createTexture();
		TextureMetadata metadata = metadataByColor(color);
		metadata.applyArguments(textureType());
		GLFunctions.upload2DImage(textureType(), 0, metadata, bounds, bounds, 0, 0L);
	}
	
	public Glyth build(IGlythSheetInfo info) {
		if(color != info.isColored()) return null;
		Slot result = slot.insert(info);
		if(result != null) {
			bind();
			info.upload(result.x, result.y);
			float minU = (slot.x + 0.01F) / bounds;
			float maxU = (slot.x - 0.01F + info.width()) / bounds;
			float minV = (slot.y + 0.01F) / bounds;
			float maxV = (slot.y - 0.01F + info.height()) / bounds;
			return new Glyth(id(), minU, minV, maxU, maxV);
		}
		return null;
	}
	
	public boolean colored() { return color; }
	@Override
	public int width() { return bounds; }
	@Override
	public int height() { return bounds; }
	
	public static class Slot {
		final int x;
		final int y;
		final int width;
		final int height;
		Slot[] children = null;
		boolean occupied;
		
		public Slot(int x, int y, int width, int height) {
			this.x = x;
			this.y = y;
			this.width = width;
			this.height = height;
		}
		
		public Slot insert(IGlythSheetInfo info) {
			if(children == null) {
				if(occupied) return null;
				int iw = info.width();
				int ih = info.height();
				if(iw > width || ih > height) return null;
				if(iw == width && ih == height) {
					occupied = true;
					return this;
				}
				int dw = width - iw;
				int dh = height - ih;
				children = new Slot[dw > 0 && dh > 0 ? 3 : 2];
				children[0] = new Slot(x, y, iw, ih);
				expandSlot(iw, ih, dw, dh);
				return children[0].insert(info);
			}
			for(int i = 0,m=children.length;i<m;i++) {
				Slot slot = children[i].insert(info);
				if(slot != null) return slot;
			}
			return null;
		}
		
		private void expandSlot(int rw, int rh, int dw, int dh) {
			if(dw > 0 && dh > 0) {
				if(dw >= dh) {
					children[1] = new Slot(x + rw, y, dw, rh);
					children[2] = new Slot(x, y + rh, width, dh);
				}
				else {
					children[1] = new Slot(x, y + rh, rw, dh);
					children[2] = new Slot(x + rw, y, dw, height);
				}
			}
			else if(dw == 0) children[1] = new Slot(x, y + rh, rw, dh);
			else if(dh == 0) children[1] = new Slot(x + rw, y, dw, rh);
		}
		
		@Override
		public String toString() {
			return "Slot[x="+x+", y="+y+", w="+width+", h="+height+"]";
		}
	}
}
